@@ -1,4 +1,4 @@
-VERSION = (1, 7, 3)
+VERSION = (1, 7, 5)
 __version__ = '.'.join(map(str, VERSION))
 
 
@@ -58,6 +58,10 @@ def ensure_completely_loaded(force=False):
             except AttributeError:
                 pass
 
+        # Randomly call some cache filling methods
+        # http://goo.gl/XNI2qz
+        model._meta._fill_fields_cache()
+
     # Calls to get_models(...) are cached by the arguments used in the call.
     # This cache is normally cleared in loading.register_models(), but we
     # invalidate the get_models() cache, by calling get_models
